from typing import Optional, TypeVar

T = TypeVar('T')

class Node:
    data: Optional[T]
    connections: list[int]
    visited: bool = ...
    def __init__(self, data: Optional[T], connections: list[int]=...) -> None: ...
    def clone(self) -> Node[T]: ...



from typing import TypeVar

T = TypeVar('T')

class Stack:
    def pop(self) -> T: ...
    def peek(self) -> T: ...
    def push(self, item: T) -> None: ...
    def isEmpty(self) -> bool: ...
    def __len__(self) -> int: ...



#!python3.9

# support type hinting in editor and code
from typing import Generator, Generic, Iterator, List, Optional, TypeVar

# circular queue for breadth first search
from datastructures.circular_queue import CircularQueue

# graph node
from datastructures.graph_node import Node

T = TypeVar("T")


class Graph(Generic[T]):
    """A binary graph of type {T}."""

    __nodes: List[Node[T]] = []  #  the list of nodes of this graph

    def __init__(self, nodes: Optional[list[Node[T]]] = None):
        """Constructor for a binary tree of type {T}.

        Args:
            type (T): The value type of the values of the nodes of the tree. For example, a tree storing names would have a value type of {str}.

        Returns:
            None.

        Instantiating a graph:
        >>> graph = Graph[str]()
        >>> graph
        []
        """

        # Check if any nodes were provided
        if nodes is not None:
            # And if any were, set the nodes to them
            self.setNodesFromNodesList(nodes)

    def __repr__(self) -> str:
        """Return a string representation of this object.

        Returns:
            str: The string representation of this object.
        """
        # map each of the nodes in the graph to a string and put all those into a list
        # and then return the stringified list.
        return str(list(map(str, self.__nodes)))

    def __iter__(self) -> Generator[Node[T], None, None]:
        """Make this class iterable

        Yields:
            Generator[Node[T], None, None]: The nodes of the graph, in the order that the graph was initialized with them.
        """
        for node in self.__nodes:
            yield node

    def __getitem__(self, key: int) -> Optional[T]:
        index = 0
        for node in self:
            if index == key:
                return node.data
            index += 1
        # not found
        raise IndexError(f"Value at index {key} not found.")

    def __len__(self) -> int:
        return len(self.__nodes)

    def getConnectionsOfNodeAtIndex(self, index: int) -> List[int]:
        """Get the connections of a node at specified index.

        Args:
            index (int): The index of the node.

        Returns:
            List[int]: The list indices of its connected cells.
        """
        return self.__nodes[index].connections

    def setNodeData(self, index: int, newValue: T) -> None:
        for nodeIndex in range(len(self)):
            if nodeIndex == index:
                # set the node data bc index is correct
                self.__nodes[nodeIndex].data = newValue
                # make sure to return so we don't loop over everything!
                return
        # not found
        raise IndexError(f"Value at index {index} not found.")

    @staticmethod
    def createGraph(sizeX: int, sizeY: int) -> "Graph[T]":
        """Creates a graph of the specified X and Y size.

        Args:
            sizeX (int): the desired X size of the graph.
            sizeY (int): the desired Y size of the graph.

        Returns:
            Graph[T]: The (empty) graph object (of the given size).

        >>> graph = Graph.createGraph(2, 2)
        >>> [i for i in map(str, graph)]
        ['None -> []', 'None -> []', 'None -> []', 'None -> []']
        """

        if (sizeX < 0) or (sizeY < 0):
            raise ValueError(f"Invalid size `({sizeX}, {sizeY})` given.")

        # make a list of nodes of the correct length
        everyNode = [Node[T](None) for _ in range(0, sizeX * sizeY)]
        return Graph(nodes=everyNode)

    def setNodesFromNodesList(self, nodes: list[Node[T]]) -> None:
        """Set a graph's nodes from a list of nodes.

        Args:
            nodes (list[Node[T]]): The list of nodes from which to set the graph data.

        >>> sampleMaze = Graph[None]()
        >>> sampleMaze.setNodesFromNodesList([
        ...     Node(None,  [1, 4]),         # 0
        ...     Node(None,  [0, 4]),         # 1
        ...     Node(None,  [4, 3]),         # 2
        ...     Node(None,  [2, 4]),         # 3
        ...     Node(None,  [0, 1, 2, 3])    # 4
        ... ])
        >>> sampleMaze
        ['None -> [1, 4]', 'None -> [0, 4]', 'None -> [4, 3]', 'None -> [2, 4]', 'None -> [0, 1, 2, 3]']
        """
        self.__nodes = nodes

    def setNodesFromValuesAndConnections(
        self, values: list[T], connectionsPointers: list[list[int]]
    ) -> None:
        """Give a Tree object some nodes.

        Args:
            values (List[T]): the value of each node.
            connectionsPointers (List[List[int]]): the connections that each node has.

        Setting a graph's nodes from provided data and connections:
        >>> maze = Graph[None]()
        >>> maze.setNodesFromValuesAndConnections([
        ...     None,
        ...     None,
        ...     None,
        ...     None,
        ...     None,
        ... ], [
        ...     [1, 4],
        ...     [0, 4],
        ...     [4, 3],
        ...     [2, 4],
        ...     [0, 1, 2, 3]
        ... ])
        >>> maze
        ['None -> [1, 4]', 'None -> [0, 4]', 'None -> [4, 3]', 'None -> [2, 4]', 'None -> [0, 1, 2, 3]']
        """

        for i in range(len(values)):
            # make new `Node` with this value's data and left and right pointers
            thisNode: Node[T] = Node(data=values[i], connections=connectionsPointers[i])
            self.__nodes.append(thisNode)  # append new node to `self.nodes`

    def connectionExistsFrom(self, indexA: int, indexB: int) -> bool:

        """Check whether there is a connection from provided node index A to index B.

        Args:
            indexA (int): The index of the `from` node – this is checked whether it has a connection to `indexB`.
            indexB (int): The index of the `to` node.

        Returns:
            bool: whether there is a link from node indexA to node indexB.

        Raises:
            IndexError: if the node at `indexA` is nonexistent.

        >>> Graph[str]([
        ...     Node('Entrance', [1]),
        ...     Node('1', [0, 2, 3]),
        ...     Node('2', [1, 4]),
        ...     Node('3', [1, 5]),
        ...     Node('4', [2]),
        ...     Node('5', [3, 6, 7]),
        ...     Node('6', [5, 8]),
        ...     Node('7', [5, 9, 10]),
        ...     Node('Exit', [6, 11]),
        ...     Node('9', [7]),
        ...     Node('10', [7]),
        ...     Node('11', [8])
        ... ]).connectionExistsFrom(0, 1)
        True

        >>> Graph[str]([
        ...     Node('Entrance', [1]),
        ...     Node('1', [0, 2, 3]),
        ...     Node('2', [1, 4]),
        ...     Node('3', [1, 5]),
        ...     Node('4', [2]),
        ...     Node('5', [3, 6, 7]),
        ...     Node('6', [5, 8]),
        ...     Node('7', [5, 9, 10]),
        ...     Node('Exit', [6, 11]),
        ...     Node('9', [7]),
        ...     Node('10', [7]),
        ...     Node('11', [8])
        ... ]).connectionExistsFrom(48, 4)
        Traceback (most recent call last):
        ...
        IndexError: Node at index 48 is nonexistent.
        """
        try:
            # check there's a connection between indexA and indexB
            return indexB in self.__nodes[indexA].connections
        except IndexError:
            raise IndexError("Node at index {} is nonexistent.".format(str(indexA)))

    def __setAllNotVisited(self) -> None:
        """Set all the nodes of the graph's `visited` status to `False`."""
        for node in self.__nodes:
            node.visited = False

    def _exists(self, nodePointer: int) -> bool:
        """Determine whether or not a pointer points to a node which exists.

        Args:
            nodePointer (int): The index of the node the pointer points to.

        Returns:
            bool: Whether the pointer is valid.

        >>> maze = Graph[None]()
        >>> maze.setNodesFromNodesList([
        ...     Node(None,  [1, 4]),         # 0
        ...     Node(None,  [0, 4]),         # 1
        ...     Node(None,  [4, 3]),         # 2
        ...     Node(None,  [2, 4]),         # 3
        ...     Node(None,  [0, 1, 2, 3])    # 4
        ... ])
        >>> maze._exists(4)
        True
        >>> maze._exists(5)
        False
        """

        try:
            _ = self.__nodes[nodePointer].data
        except IndexError:
            # index out of range for `self.__nodes`
            return False
        else:
            # the index is OK
            return True

    def depthFirstTraversal(self, nodeIndex: Optional[int] = None) -> Iterator[T]:
        """Recursively depth-first traverse the graph.
        Yields data — rather than returning it — because returning the values would require putting everything into a list.

        `yield` basically pauses the function after yielding, saving all of its states, and only
        when the next value is required by the caller function does the yielding function continue.

        Complexity of this implementation is `O(V * E)` where `V` is the number of vertices (nodes) and `E` is the number of edges (connections).

        Args:
            currentNode (Optional[int], optional): Specify the node index to start from. Used for recursion; don't worry about this parameter. Defaults to None.

        Raises:
            Exception: The graph is empty, and therefore we cannot traverse it.

        Yields:
            Generator[Node[T]]: The nodes' data in depth-first order.

        >>> sampleGraph = Graph[int]([
        ...     Node(2,  [1, 4]),         # 0
        ...     Node(8,  [0, 4]),         # 1
        ...     Node(42,  [4, 3]),         # 2
        ...     Node(11,  [2, 4]),         # 3
        ...     Node(91,  [0, 1, 2, 3])    # 4
        ... ])
        >>> [item for item in sampleGraph.depthFirstTraversal()]
        [11, 42, 91, 8, 2]

        >>> [item for item in Graph[int]().depthFirstTraversal()]       # test traversal of empty graph
        []
        """

        # check there's actually any nodes
        if len(self.__nodes) < 1:
            # there's no nodes.
            return

        if nodeIndex is None:
            # the function has been called without a starting node so we start from the beginning!
            # so we'll make everything not visited to be able to visit stuff.
            self.__setAllNotVisited()
            # check that there are actually any items in the graph
            if (self.__nodes is []) or (self.__nodes is None):
                # there are no nodes in the graph...
                raise Exception("Cannot traverse an empty graph")
            else:
                # there are nodes in the graph :)
                # ...so we set `nodeIndex` to the first index:
                nodeIndex = 0

        # `nodeIndex` is set and guaranteed to have a value at this point
        # start with node `self.__nodes[nodeIndex]`
        # only continue if `self.__nodes[nodeIndex]` is not visited
        if self.__nodes[nodeIndex].visited == False:
            # the node is not visited
            # so set it as visited
            self.__nodes[nodeIndex].visited = True
            # for each of this node's neighbours
            for connectionIndex in self.__nodes[nodeIndex].connections:
                # perform a depth first traversal of the neighbour
                # `yield from` (emphasis on the `from`) because it is yielding the result of a recursive yield
                # see PEP 380 Syntax for Delegating to a Subgenerator – https://www.python.org/dev/peps/pep-0380/
                yield from self.depthFirstTraversal(connectionIndex)

            # check the data is not None and then yield it
            data = self.__nodes[nodeIndex].data
            if data is not None:
                yield data  # this was previously guaranteed to be a safe index

    def breadthFirstTraversal(self) -> Iterator[Optional[T]]:
        """Iteratively breadth-first traverse the graph.
        Yields data – rather than returning it – because returning values would require more memory.
        `yield` basically pauses the function after yielding, saving all of its states, and only
        when the next value is required by the caller function does the yielding function continue.

        Yields:
            Generator[Node[T]]: The nodes' data in depth-first order.

        >>> simply_connected_maze = Graph[str]([
        ...     Node('Entrance', [1]),
        ...     Node('1', [0, 2, 3]),
        ...     Node('2', [1, 4]),
        ...     Node('3', [1, 5]),
        ...     Node('4', [2]),
        ...     Node('5', [3, 6, 7]),
        ...     Node('6', [5, 8]),
        ...     Node('7', [5, 9, 10]),
        ...     Node('Exit', [6, 11]),
        ...     Node('9', [7]),
        ...     Node('10', [7]),
        ...     Node('11', [8])
        ... ])
        >>> [maze_cell for maze_cell in simply_connected_maze.breadthFirstTraversal()]  # test traversal of a simple connected maze
        ['Entrance', '1', '2', '3', '4', '5', '6', '7', 'Exit', '9', '10', '11']

        >>> non_simple_maze = Graph[str]([
        ...     Node('Entrance', [1, 2]),
        ...     Node('1', [0, 3, 4]),
        ...     Node('2', [0, 5]),
        ...     Node('3', [1, 6]),
        ...     Node('4', [1, 7]),
        ...     Node('5', [7, 8]),
        ...     Node('6', [3, 9]),
        ...     Node('7', [4, 5, 9, 10]),
        ...     Node('8', [5]),
        ...     Node('9', [6, 7, 11]),
        ...     Node('10', [7]),
        ...     Node('Exit', [9])
        ... ])
        >>> [maze_cell for maze_cell in non_simple_maze.breadthFirstTraversal()]
        ['Entrance', '1', '2', '3', '4', '5', '6', '7', '8', '9', '10', 'Exit']

        >>> [item for item in Graph[int]().breadthFirstTraversal()]     # test a traversal of an empty graph
        []
        """

        # check there's actually any nodes
        if len(self.__nodes) < 1:
            # there's no nodes.
            return

        # initialize a queue for the visited nodes
        visitedNodes = CircularQueue[Node[T]](len(self.__nodes))

        # set the 0th node as visited
        self.__nodes[0].visited = True
        # and enqueue it
        visitedNodes.enQueue(self.__nodes[0])

        # for each item in the queue:
        while len(visitedNodes) > 0:
            # pop a node from the queue
            currentNode = visitedNodes.deQueue()
            # and yield it
            yield currentNode.data  # we previously guaranteed that this value is safe with the while statement

            # get neighbours of the node
            for neighbour in currentNode.connections:
                # check if it's been visited
                if self.__nodes[neighbour].visited == False:
                    # set it to visited bc we're visiting it now
                    self.__nodes[neighbour].visited = True
                    # and add it to the `visitedNodes` queue
                    visitedNodes.enQueue(self.__nodes[neighbour])

    def __checkIndexIsValidWithException(self, index: int) -> bool:
        if not (len(self.__nodes) > index >= 0):
            raise IndexError("Node at index {} is out of range.".format(str(index)))
        return True

    def removeLinkBetween(
        self,
        indexFrom: int,
        indexTo: int,
        bidirectional: bool = True,
    ) -> None:
        """Removes a link between two nodes

        Args:
            indexFrom (int): The `from` node to connect to the `to`
            indexTo (int): The `to` node
            bidirectional (bool, optional): Whether or not to remove the link both ways. Defaults to True.
        """

        # make sure indices are valid
        self.__checkIndexIsValidWithException(indexFrom)
        self.__checkIndexIsValidWithException(indexTo)

        # check it's in the list of indices
        if indexTo in self.__nodes[indexFrom].connections:
            self.__nodes[indexFrom].connections.remove(indexTo)
        else:
            # it isn't
            raise ValueError(
                f"Node index {indexTo} already does not exist in node at index {indexFrom}'s connections.",
            )

        # if bidirectional, flip indexTo and indexFrom and do it again
        if bidirectional:
            self.removeLinkBetween(indexTo, indexFrom, False)

    def addLinkBetween(
        self,
        indexFrom: int,
        indexTo: int,
        bidirectional: bool = True,
    ) -> None:
        """Add a link between two nodes of given indices.

        Args:
            indexFrom (int): the 'from' node to connect to the 'to'
            indexTo (int): the 'to' node
            bidirectional (bool, optional): Whether or not to connect the link both ways (i.e., `indexFrom` to `indexTo` and vice versa.). Defaults to True.

        Raises:
            Exception: If the node index `indexFrom` already exists in `indexTo`'s connections.

        >>> simply_connected_maze = Graph[str]([
        ...     Node('Entrance', [1]),
        ...     Node('1', [0, 2, 3]),
        ...     Node('2', [1, 4]),
        ...     Node('3', [1, 5]),
        ...     Node('4', [2]),
        ...     Node('5', [3, 6, 7]),
        ...     Node('6', [5, 8]),
        ...     Node('7', [5, 9, 10]),
        ...     Node('Exit', [6, 11]),
        ...     Node('9', [7]),
        ...     Node('10', [7]),
        ...     Node('11', [8])
        ... ])

        >>> simply_connected_maze.addLinkBetween(2, 3)

        >>> simply_connected_maze.addLinkBetween(2, 4)
        Traceback (most recent call last):
        ...
        ValueError: Node index '4' already exists in node 2's connections.

        >>> simply_connected_maze.addLinkBetween(55, 3)
        Traceback (most recent call last):
        ...
        IndexError: Node at index 55 is out of range.

        >>> simply_connected_maze.addLinkBetween(2, 333)
        Traceback (most recent call last):
        ...
        IndexError: Node at index 333 is out of range.
        """

        # make sure indices are valid
        self.__checkIndexIsValidWithException(indexFrom)
        self.__checkIndexIsValidWithException(indexTo)

        # check the node isn't already connected
        if self.connectionExistsFrom(indexFrom, indexTo):
            # it is, so raise an exception with a descriptive error message
            raise ValueError(
                "Node index '{}' already exists in node {}'s connections.".format(
                    str(indexTo), str(indexFrom)
                )
            )
        else:
            # add the node index in its indexA's connections list

            # do this by using a temporary variable to avoid errors when mutating list values' values
            nodeTemp: Node[T]

            try:
                nodeTemp = self.__nodes[indexFrom].clone()
            except IndexError:
                raise IndexError(
                    "Node at index {} is nonexistent.".format(str(indexFrom))
                )

            # set our temporary vars correctly
            nodeTemp.connections.append(indexTo)
            # and then set the node we want to the correctly set temporary variable
            self.__nodes[indexFrom] = nodeTemp

        if bidirectional:
            # do it again!
            #  but make sure to not do it bidirectionally because then it'd go on forever
            self.addLinkBetween(indexTo, indexFrom, False)



from datastructures._queue_position import QueuePosition as QueuePosition
from datastructures.linked_list import LinkedList as LinkedList
from typing import TypeVar

T = TypeVar('T')

class CircularQueue:
    def __init__(self, maxSize: int) -> None: ...
    def __len__(self) -> int: ...
    def enQueue(self, item: T) -> None: ...
    def deQueue(self) -> T: ...
    def get(self, position: QueuePosition) -> T: ...
    def isEmpty(self) -> bool: ...
    def isFull(self) -> bool: ...



#!python3.9

# queue position enum for specifying the position in the queue
from datastructures._queue_position import QueuePosition

# linked list for enhanced queue
from datastructures.linked_list import LinkedList

# type hinting and some safety
from typing import Optional, TypeVar, Generic

T = TypeVar("T")


class CircularQueue(Generic[T]):
    """A queue object – first in, first out data structure designed in this case around readability, maintainability and efficiency.

    Args:
        Generic ([type]): The type of data this queue will store.

    Create a circular queue and test its length and maxSize
    >>> myQueue = CircularQueue[float](100)
    >>> len(myQueue)
    0

    Check maximum size of queue
    >>> myQueue._CircularQueue__maxSize     # accessing private attribute for testing
    100
    """

    # this is a circular queue. It is implemented as such
    __queue: LinkedList[Optional[T]]
    __maxSize: int
    __length: int
    __tail: int
    __head: int

    def __init__(self, maxSize: int):
        """Initialize an empty queue with maximum size.

        Args:
            maxSize (int): The maximum size of the queue.

        Initializing of a queue:
        >>> myQueue = CircularQueue[str](6)
        ... # Empty queue has correct length and initial items:
        >>> print(myQueue._CircularQueue__queue)       # accessing private attr for testing
        [None, None, None, None, None, None]

        Check the maximum size of the queue
        >>> myQueue._CircularQueue__maxSize     # accessing private attr for testing
        6
        """
        # initialize `self.queue` to an array of `-1` repeating `maxSize` number of times
        self.__queue = LinkedList[Optional[T]]([None] * maxSize)  # type: ignore     # for type hinting to work correctly
        # set `self._maxSize` to provided `maxSize` argument
        self.__maxSize = maxSize
        # set `self.__length` to 0, because there are no elements in our queue just yet
        self.__length = 0
        # and set both `self._tail` and `self._head` to `-1` as they don't point to anything yet
        self.__tail = -1
        self.__head = -1

    def __len__(self) -> int:
        return self.__length

    def enQueue(self, item: T) -> None:
        """Enqueue an item to the queue. Time complexity is O(1).

        Args:
            item (T): The item of type T to enqueue.

        Raises:
            Exception: Queue is full – it has reached the provided maximum size.

        Initialize a queue and add an acceptable number of items to it:
        >>> myQueue = CircularQueue[str](6)
        >>> myQueue.enQueue("Hello")
        >>> myQueue.enQueue("there,")
        >>> myQueue.enQueue("I")
        >>> myQueue.enQueue("am")
        >>> myQueue.enQueue("an")
        >>> myQueue.enQueue("octopus.")

        Check queue is full:
        >>> myQueue.isFull()
        True

        Try to add too many items:
        >>> myQueue.enQueue("overflow!")
        Traceback (most recent call last):
            ...
        Exception: Queue is full

        """

        # check if queue is full
        if not self.isFull():
            """
                The range for the head and tail pointer should be between 0 and `maxSize - 1`,
            hence we are using the logic that if we divide x by 5, then the remainder can
            never be greater than 5. In other words, it should be between 0 and 4. So apply
            this logic to the formulae:
                tail = (tail + 1) % maxSize
                head = (head + 1) % maxSize

                Observe that this helps us to avoid reinitializing tail and head to 0 when
            the queue becomes full.
            """
            self.__tail = (self.__tail + 1) % self.__maxSize
            self.__queue[self.__tail] = item
            self.__length += 1
            if self.__length == 1:
                self.__head = 0
        else:
            # queue is full! uh oh...
            raise Exception("Queue is full")

    def deQueue(self) -> T:
        """Dequeue the last item in the queue and return it. Time complexity is O(1).

        Raises:
            Exception: The queue is empty, so you can't dequeue anything.

        Returns:
            T: The dequeued item.

        Initialize a queue, add stuff, dequeue everything and test it all returns:
        >>> myQueue = CircularQueue[str](6)
        >>> myQueue.enQueue("Hello")
        >>> myQueue.enQueue("there,")
        >>> myQueue.enQueue("I")
        >>> myQueue.enQueue("am")
        >>> myQueue.enQueue("an")
        >>> myQueue.enQueue("octopus.")
        >>> # Now dequeue everything:
        >>> allTheElementsOfTheQueue = list[str]()
        >>> while myQueue.isEmpty() == False:
        ...     # dequeue an element...
        ...     dequeuedElement = myQueue.deQueue()
        ...     # ...and then print it. It is safe to print because the element is of type `str` (as we initialized our queue with the same element type).
        ...     allTheElementsOfTheQueue.append(dequeuedElement)
        >>> print(' '.join(allTheElementsOfTheQueue))
        Hello there, I am an octopus.

        The length of the queue should now be 0 as we dequeued everything:
        >>> len(myQueue)
        0
        """

        if self.isEmpty():
            # can't dequeue anything from an empty queue!
            raise Exception("Queue empty")

        # store in a temporary variable because if we return the value here, the rest of the function does not execute
        _dataToReturn = self.__queue[self.__head]

        self.__head = (self.__head + 1) % self.__maxSize
        # decrement `self.__length` by 1
        self.__length -= 1

        if self.isEmpty():
            # set head and tail to `-1` if the queue is empty
            self.__head = -1
            self.__tail = -1

        if _dataToReturn is not None:
            return _dataToReturn  # we previously guaranteed the list to not be empty with `self.isEmpty()``
        else:
            raise RuntimeError(
                "Queue was modified outside of package scope - datastructures/circular_queue/circular_queue.py:171"
            )

    def get(self, position: QueuePosition) -> T:
        """Fetch an item from the front or the back of the queue without removing it from the queue

        Args:
            position (QueuePosition): The front or back of the queue – `QueuePosition.front` or `QueuePosition.back`

        Raises:
            Exception: Queue is empty and therefore cannot fetch any element from it
            Exception: Invalid QueuePosition given

        Returns:
            T: The fetched item

        Add some items to a queue and fetch the front item:
        >>> myQueue = CircularQueue[str](2)
        >>> myQueue.enQueue("Hello")
        >>> myQueue.enQueue("there")
        >>> myQueue.get(QueuePosition.front)
        'Hello'

        Fetch the rear item:
        >>> myQueue.get(QueuePosition.rear)
        'there'

        Add one item to a queue of length 1 and fetch the front:
        >>> myTinyQueue = CircularQueue[float](1)
        >>> myTinyQueue.enQueue(3.14159265)
        >>> myTinyQueue.get(QueuePosition.front)
        3.14159265

        Fetch the rear item from the 1-length queue:
        >>> myTinyQueue.get(QueuePosition.rear)
        3.14159265

        Add only one item to a queue of length 5 and fetching the front item:
        >>> myMediumQueue = CircularQueue[int](5)
        >>> myMediumQueue.enQueue(2)
        >>> myMediumQueue.get(QueuePosition.front)
        2

        Fetch the rear item from the 5-length queue with 1 item in:
        >>> myMediumQueue.get(QueuePosition.rear)
        2

        Using an invalid QueuePosition:
        >>> myQueue.get("hula hoops")
        Traceback (most recent call last):
            ...
        Exception: Queue position must be 'front' or 'rear'
        """
        if self.isEmpty():
            raise Exception("Queue is empty")

        if position == QueuePosition.front:
            return self.__queue[
                self.__head
            ]  # type:ignore   # already made sure the queue is not empty
        elif position == QueuePosition.rear:
            return self.__queue[
                self.__tail
            ]  # type:ignore   # again, it's guaranteed to be safe

        raise Exception("Queue position must be 'front' or 'rear'")

    def isEmpty(self) -> bool:
        """Check whether the queue is empty.

        Returns:
            bool: Whether or not the queue is empty.

        Initialize a queue and test if it is empty:
        >>> myQueue = CircularQueue[str](1)
        >>> myQueue.isEmpty()
        True
        >>> myQueue.enQueue("Foo")
        >>> myQueue.isEmpty()
        False
        """
        return self.__length == 0

    def isFull(self) -> bool:
        """Check whether the queue is full.

        Returns:
            bool: Whether or not the queue is full.

        Initialize a queue and check if the queue is full:
        >>> myQueue = CircularQueue[str](1)
        >>> myQueue.isFull()
        False

        Make it full and check it is now full:
        >>> myQueue.enQueue("octopus.")
        >>> myQueue.isFull()
        True
        """
        return self.__length == self.__maxSize



from enum import Enum

class QueuePosition(Enum):
    """An enum for either the front or rear of a queue

    Args:
        Enum ([type]): Automatically provided by the enum module.

    Example usage:
        `QueuePosition.front`
        `QueuePosition.rear`

    Check QueuePosition initializes properly
    >>> front = QueuePosition.front
    >>> print(front)
    QueuePosition.front
    >>> rear = QueuePosition.rear
    >>> print(rear)
    QueuePosition.rear
    """

    front = 0
    rear = 1



from typing import Optional, TypeVar

T = TypeVar('T')

class LinkedListNode:
    data: T
    nextNode: Optional[LinkedListNode[T]]
    def __init__(self, data: T, nextNode: Optional[LinkedListNode[T]]=...) -> None: ...
    def clone(self) -> LinkedListNode[T]: ...



#!python3.9

from typing import Generator, Generic, Optional, TypeVar

# Linked List Node for the list
from datastructures.linked_list_node import LinkedListNode

T = TypeVar("T")


class LinkedList(Generic[T]):
    __head: Optional[LinkedListNode[T]] = None

    def __init__(self, nodes: list[T] = []) -> None:
        """The constructor for a `LinkedList` object.

        Args:
            nodes (list[LinkedListNode[T]]): A list of data for the linked list to create itself from.

        Raises:
            ValueError: `nodes` argument not supplied.

        Instantiate and print an empty linked list
        >>> emptyLinkedList = LinkedList[int]()
        >>> str(emptyLinkedList)
        '[]'

        Instantiate and print a linked list of sample objects
        >>> mazeCellsIndices = LinkedList[int]([0, 1, 2, 3, 4, 5])
        >>> mazeCellsIndices._LinkedList__head
        0 (1 (2 (3 (4 (5 (None))))))
        """
        # set `self.__head` to `None` as no items have been added to the list yet
        self.__head = None
        # check that some items have been provided in this function's arguments
        if nodes is None:
            raise ValueError("`nodes` argument must be a list of `T`.")
        else:
            if len(nodes) == 0:
                # an empty list of nodes was given so initialize `self.__head` to `None` and return
                self.__head = None
                return  # don't execute the rest of the function

            # set `node` to the first item of the provided nodes
            node = LinkedListNode(nodes.pop(0))
            #  set `self.__head` to the above-defined variable `node`
            self.__head = node
            # loop over each of the nodes in the nodes provided in this function's arguments
            for thisNode in nodes:
                # set the current node's next node to a new `LinkedListNode` object of `thisNode`
                node.nextNode = LinkedListNode(thisNode)
                #  set the current node to the next node to move on
                node = node.nextNode

    def __str__(self) -> str:
        return str(self.toList())

    def __iter__(self) -> Generator[LinkedListNode[T], None, None]:
        """Helper method to iterate over items in the list. Yields each value (rather than return everything as a list) as to not use up potentially infinite memory.
        The `yield` statement pauses the function, saving all its states and later continues from there on successive calls.

        Yields:
            LinkedListNode: The next node in the linked list — either element 0 of the list, or the `nextNode` attribute of the previous node.

        Iterate over a linked list
        Instantiate a `LinkedList` and print it all out, testing iteration and generator
        >>> mazeCellsIndices = LinkedList[int]([9, 4, 3, 6])
        >>> [listItem.data for listItem in mazeCellsIndices]
        [9, 4, 3, 6]

        Iterate over an empty linked list
        >>> [item.data for item in LinkedList[str]()]
        []
        """
        # check that there are actually any items in the list
        if self.__head is not None:
            # set node to an optional `LinkedListNode` of type `T` so we can set it to the next node, which may or may not be optional. The optional allows us to check the next node exists and not accidentally yield a `None` value.
            node: Optional[LinkedListNode[T]] = self.__head
            # check the (next) node exists
            while node is not None:
                # yield the node, pausing the function, saving its states and allowing it to continue from here on successive calls.
                yield node
                # set node to the next node referenced by this node. This may or may not exist, hence we have to be careful and unwrap the value as we have above.
                node = node.nextNode

    def __getitem__(self, key: int) -> T:
        """Get item at specified index

        Args:
            key (int): the index of the item to get.

        Raises:
            IndexError: Index out of range.

        Returns:
            T: The data of the item at specified index.

        >>> mazeCellsIndices = LinkedList[int]([14, 11, 3, 6, 8])
        >>> mazeCellsIndices[0]
        14
        >>> mazeCellsIndices[2]
        3
        >>> mazeCellsIndices[4]
        8
        >>> mazeCellsIndices[6] # doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
        ...
        IndexError: LinkedList index 6 is out of range
        >>> mazeCellsIndices[-1] # doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
        ...
        IndexError: LinkedList index -1 is out of range
        """
        index = 0
        currentNode = self.__head

        while currentNode is not None:
            if index == key:
                return currentNode.data
            elif index > key:
                raise IndexError(
                    "LinkedList index {} is out of range".format(str(index))
                )
            else:
                # we're not there yet so increment index and set node to the nextNode
                index += 1
                currentNode = currentNode.nextNode
        # looped over everything without finding index
        raise IndexError("LinkedList index {} is out of range".format(str(index)))

    def __setitem__(self, key: int, newValue: T):
        """setitem method of LinkedList. Used to set an item specified at an index in the LinkedList.

        Args:
            key (int): The index of the LinkedList item to set.
            newValue (T): The new value of the LinkedList item.

        Raises:
            IndexError: Index of `key` out of range.

        Test setting linked list nodes via their indices
        >>> indices = LinkedList[int]([1, 2, 3, 4])
        >>> indices[0] = 5
        >>> indices[1] = 6
        >>> indices[2] = 7
        >>> indices[3] = 8
        >>> print(indices)
        [5, 6, 7, 8]

        >>> indices[4] = 9
        >>> print(indices)
        [5, 6, 7, 8, 9]

        >>> indices[6] = 11
        Traceback (most recent call last):
        ...
        IndexError: LinkedList index 6 is out of range

        >>> indices[-1] = 4
        Traceback (most recent call last):
        ...
        IndexError: LinkedList index -1 is out of range
        """
        self.setItemRecursively(key, newValue)

    def setItemRecursively(
        self,
        index: int,
        newValue: T,
        currentNode: Optional[LinkedListNode[T]] = None,
        recursiveIndex: int = 0,
    ) -> Optional[LinkedListNode[T]]:
        """Recursively set a node's data at specified index.

        Args:
            index (int): The index of the node whose data you wish to set.
            newValue (T): The new value for the node's data
            currentNode (Optional[LinkedListNode[T]], optional): Used for recursion, keeping track of nodes. Defaults to None.
            recursiveIndex (int, optional): Used for recursion, the value for the index of the recursion. Defaults to 0.

        Raises:
            IndexError: LinkedList index is out of range.

        Returns:
            Optional[LinkedListNode[T]]: Used for recursion. The updated LinkedListNode.
        """

        if (recursiveIndex == 0) and (currentNode is None):
            # if this function was called with default recursiveIndex and currentNode parameters
            currentNode = self.__head

        if currentNode is None:
            if index > recursiveIndex:
                # the index is out of range...
                raise IndexError(
                    "LinkedList index {} is out of range".format(str(index))
                )

        if recursiveIndex == index:
            # we've got to the right node index
            # check if this node is None so we can set nextNode appropriately
            if currentNode is None:
                nextNode = None
            else:
                # this node is not None so we just set nextNode to this node's nextNode
                nextNode = currentNode.nextNode
            # we've set nextNode to appropriate value here so we make a new node and return it
            newNode = LinkedListNode[T](newValue, nextNode)
            self.__head = newNode
            return newNode
        elif recursiveIndex < index:
            # we're not there yet
            if currentNode is None:
                # the index is out of range
                raise IndexError(
                    "LinkedList index {} is out of range".format(str(index))
                )
            else:
                # call recursively to get the nextNode
                nextNode = self.setItemRecursively(
                    index, newValue, currentNode.nextNode, recursiveIndex + 1
                )
                # set newNode to this node with the previously set nextNode as its nextNode
                newNode = LinkedListNode[T](currentNode.data, nextNode=nextNode)
                self.__head = newNode
                return newNode
        elif recursiveIndex > index:
            # gone over!
            raise IndexError("LinkedList index {} is out of range".format(str(index)))

    def toNode(self) -> Optional[LinkedListNode[T]]:
        return self.__head

    def clone(self) -> "LinkedList[T]":
        """Clone this class object to overcome Python's pass-by-reference argument types.

        Returns:
            LinkedList[T]: clone of the object.
        """
        linkedList = LinkedList[T]()
        if self.__head is not None:
            linkedList.extendByNode(self.__head)
        return linkedList

    def toList(self) -> list[T]:
        """Generate a list of the data of all the objects in the linked list.

        Returns:
            list[T]: The data of each of the linked list nodes.

        Create Python list from instantiated linked list
        >>> mazeCellsIndices = LinkedList[int]([11, 13, 17, 19])
        >>> mazeCellsIndices.toList()
        [11, 13, 17, 19]

        Create Python list from empty linked list
        >>> emptyList = LinkedList[str]()
        >>> emptyList.toList()
        []
        """
        return [item.data for item in self]

    def insertLinkedListAtBeginning(self, linkedList: "LinkedList[T]") -> None:
        """Insert a linked list at the start of another.

        Args:
            linkedList (LinkedList): The linked list to insert at the beginning of `self`.

        Insert linked list at beginning of another
        >>> mazeCellsIndices = LinkedList[int]([11, 13, 17, 19])
        >>> moreMazeCellsIndices = LinkedList[int]([2, 3, 5, 7])
        >>> mazeCellsIndices.insertLinkedListAtBeginning(moreMazeCellsIndices)
        >>> print(mazeCellsIndices)
        [2, 3, 5, 7, 11, 13, 17, 19]
        >>> print(moreMazeCellsIndices)
        [2, 3, 5, 7]

        Insert empty list at beginning of another
        >>> moreMazeCellsIndices.insertLinkedListAtBeginning(LinkedList[int]())
        >>> print(moreMazeCellsIndices)
        [2, 3, 5, 7]

        Insert linked list at beginning of empty linked list
        >>> evenMoreMazeCellsIndices = LinkedList[int]([1, 4])
        >>> emptyLinkedList = LinkedList[int]()
        >>> emptyLinkedList.insertLinkedListAtBeginning(evenMoreMazeCellsIndices)
        >>> print(emptyLinkedList)
        [1, 4]

        Insert empty linked list at beginning of empty linked list
        >>> emptyList = LinkedList[int]()
        >>> emptyList.insertLinkedListAtBeginning(LinkedList[int]())
        >>> print(emptyList)
        []
        """

        node = linkedList.toNode()

        if node is not None:
            self.insertNodeAtBeginning(node)

    def insertNodeAtBeginning(
        self, nodeValues: LinkedListNode[T], clone: bool = True
    ) -> None:
        """Recursively inserts any number of nodes in the beginning of this LinkedList by checking if each provided node's `nextNode` attribute is set.

        Args:
            node (LinkedListNode[T]): The node to add potentially with a nextNode attribute, which may contain a node with a nextNode attribute, which may contain a node with a nextNode attribute, which may contain a node with a nextNode attribute, which...
            clone (bool): whether to clone the passed values. Only used when recursing or for special cases where you want the original list edited too.
        """

        if clone:
            # then don't modify the passed-in LinkedListNode argument
            node = nodeValues.clone()  # to pass by value, not reference
        else:
            # recursively called so we want to modify the passed-in LinkedList argument
            node = nodeValues

        # check if this node has a `nextNode`, and if so, call this same function on the `nextNode`, again and again, until there's no `nextNode`...
        if node.nextNode is None:
            # `node.nextNode` is None
            # this node's `nextNode` is None so we can set its `nextNode` to `self.__head`... -->
            node.nextNode = self.__head
        else:
            # `node.nextNode` is *not* None
            #  so we call this function recursively to traverse the chain of `LinkedListNodes` to be able to find the last node in the list and add them to our list.
            self.insertNodeAtBeginning(
                node.nextNode, clone=False
            )  # modifies the inout node parameter
            # --> ...and after we have gotten to setting the node (whose `nextNode` is `None`) to `self.__head`, we then set our `self.__head` to the node to have finished inserting the nodes.
            self.__head = node

    def extendByLinkedList(self, linkedList: "LinkedList[T]") -> None:
        """Appends the contents of a linked list to another.

        Args:
            linkedList (LinkedList[T]): The linked list to add onto `self`.

        Extend linked list by another
        >>> mazeCellsIndices = LinkedList[int]([2, 4, 12, 15])
        >>> moreMazeCellsIndices = LinkedList[int]([17, 24, 28, 31])
        >>> mazeCellsIndices.extendByLinkedList(moreMazeCellsIndices)
        >>> print(mazeCellsIndices)
        [2, 4, 12, 15, 17, 24, 28, 31]

        Extend empty linked list by a normal linked list
        >>> mazeCellsIndices = LinkedList[int]()
        >>> moreMazeCellsIndices = LinkedList[int]([1, 3, 5])
        >>> mazeCellsIndices.extendByLinkedList(moreMazeCellsIndices)
        >>> print(mazeCellsIndices)
        [1, 3, 5]

        Extend normal linked list by an empty one
        >>> mazeCellsIndices = LinkedList[int]([1, 3, 5])
        >>> moreMazeCellsIndices = LinkedList[int]()
        >>> mazeCellsIndices.extendByLinkedList(moreMazeCellsIndices)
        >>> print(mazeCellsIndices)
        [1, 3, 5]

        Extend empty linked list by an empty linked list
        >>> emptyList = LinkedList[int]()
        >>> anotherEmptyList = LinkedList[int]()
        >>> emptyList.extendByLinkedList(anotherEmptyList)
        >>> print(emptyList)
        []
        """

        node = linkedList.toNode()

        if node is not None:
            self.extendByNode(node)

    def extendByNode(
        self, node: LinkedListNode[T], insertOnto: Optional[LinkedListNode[T]] = None
    ) -> None:
        """Recursively appends nodes to the end of this LinkedList by checking each of this LinkedList's node's `nextNode` attribute is set.

        Args:
            node (LinkedListNode[T]): The node or chain of nodes you wish to append to the LinkedList.
            insertOnto (Optional[LinkedListNode[T]]): The node or chain of nodes to insert `node` onto the end of. Don't worry about this if you simply want to insert onto the end of the `LinkedList` because it traverses that automatically. Defaults to `self.__head`.
        """
        if insertOnto is None:
            # note that this function is being called without an `insertOnto` parameter, meaning we want to insert it onto the end of the `self.__head` node chain.
            if self.__head is None:
                # `self.__head` is None so we can simply set it to the given node.
                self.__head = node
            else:
                self.extendByNode(node, insertOnto=self.__head)
        # `insertOnto` is not None >
        else:
            # > meaning we have recursively called this function and intend to chain our inserts from one node to its `nextNode`.
            if insertOnto.nextNode is not None:
                # call this function (recursively) with the `nextNode` of the `insertOnto` argument of this function call.
                self.extendByNode(node, insertOnto.nextNode)
            else:
                # `insertOnto.nextNode` is None
                # We've traversed the entire chain and got to the end, where `nextNode` has no value.
                insertOnto.nextNode = node

    def removeNode(self, toRemove: T, node: Optional[LinkedListNode[T]] = None) -> None:
        """Remove a node with data `toRemove` from the list, by linking the node before it to the node after it.

        Args:
            toRemove (T): The data of the node to remove. The first node with this data will be the one removed.
            node (Optional[LinkedListNode[T]], optional): Parameter used for recursively calling this function –
            don't worry about this if you plan on starting at the start of this LinkedList. Defaults to None.

        Raises:
            Exception: List is empty, therefore the function cannot remove any value.
            Exception: The provided data was not found in any node.

        Removing a node at the start of the list
        >>> mazeCells = LinkedList[int]([2, 8, 4, 7, 16, 22, 12])
        >>> mazeCells.removeNode(2)
        >>> print(mazeCells)
        [8, 4, 7, 16, 22, 12]

        Removing a node in the middle of the list
        >>> mazeCells = LinkedList[int]([2, 8, 4, 7, 16, 22, 12])
        >>> mazeCells.removeNode(7)
        >>> print(mazeCells)
        [2, 8, 4, 16, 22, 12]

        Removing a node at the end of the list
        >>> mazeCells = LinkedList[int]([2, 8, 4, 7, 16, 22, 12])
        >>> mazeCells.removeNode(12)
        >>> print(mazeCells)
        [2, 8, 4, 7, 16, 22]

        Removing a node that does not exist
        >>> mazeCells = LinkedList[int]([2, 8, 4, 7, 16, 22, 12])
        >>> mazeCells.removeNode(5) # doctest: +IGNORE_EXCEPTION_DETAIL
        Traceback (most recent call last):
        ...
        Exception: Node not found with data '5'.
        >>> print(mazeCells)
        [2, 8, 4, 7, 16, 22, 12]

        """

        # check the specified node is `None` to see if we are starting at the beginning of this LinkedList or from a different node
        if node is None:
            # set `node` to `self.__head` to traverse the list from the beginning
            if self.__head is not None:
                node = self.__head
                # if the start of the list is the one we want to remove, just set `self.head` to the next one
                if node.data == toRemove:
                    self.__head = node.nextNode
                    return
            else:
                raise Exception("List is empty.")

        # node has been set to `self.__head`
        # check that this node's nextNode is the right one
        if node.nextNode is None:
            # we have traversed the entire list without finding the desired item
            raise Exception("Node not found with data '{}'.".format(toRemove))

        if node.nextNode.data == toRemove:
            # the next node is the one we must remove.
            # set this node's `nextNode` to its `nextNode`'s `nextNode`:
            node.nextNode = node.nextNode.nextNode
            # we are making the list 'skip out' the next node and instead pointing it to the one after that.
        else:
            # the next node is not the wanted one just yet, so we recursively call this same function with the `node` parameter as this node's `nextNode`.
            self.removeNode(toRemove, node.nextNode)



from typing import Generic, TypeVar, Optional

T = TypeVar("T")


class LinkedListNode(Generic[T]):
    """A node for a linked list object

    Args:
        Generic ([type]): The type of data this node will store
    """

    data: T  # this node's data
    # the pointer of the next node  # variable type is in quotes because we are using forward referencing here to avoid recursive typing
    nextNode: Optional["LinkedListNode[T]"]

    def __init__(self, data: T, nextNode: Optional["LinkedListNode[T]"] = None) -> None:
        """Constructor for a `LinkedListNode`

        Args:
            data (T): The data for this node to store
            next (int): The integer pointer of the next `LinkedListNode`

        Instantiating a `LinkedListNode`:
        >>> myLinkedListNode = LinkedListNode[int](65536, nextNode = 5)     # instantiate a LinkedListNode object
        >>> myLinkedListNode.data
        65536
        >>> myLinkedListNode.nextNode
        5

        """
        self.data = data
        self.nextNode = nextNode

    def __repr__(self) -> str:
        """Generate a string representation of this object
        For example, "'Barry' -> 4" for an object where `data` = 'Barry' and `nextPointer` = 4

        Returns:
            str: the string representation of the object

        >>> LinkedListNode[str](data = "Gareth", nextNode = 2)
        Gareth (2)
        """
        return "{} ({})".format(str(self.data), str(self.nextNode))

    def clone(self) -> "LinkedListNode[T]":
        """Recursively clone an object to overcome Python's pass-by-reference argument types.
        Basically used to force Python to pass by value rather than reference.

        Returns:
            LinkedListNode[T]: this object, cloned.
        """
        node = LinkedListNode[T](self.data)

        if self.nextNode is not None:
            # call this recursively to get all the values for the nodes while not copying by reference
            node.nextNode = self.nextNode.clone()
        else:
            #  got to the end of the chain of nodes
            node.nextNode = None

        return node


from datastructures.circular_queue import CircularQueue as CircularQueue
from datastructures.graph_node import Node as Node
from typing import Generator, Iterator, List, Optional, TypeVar

T = TypeVar('T')

class Graph:
    def __init__(self, nodes: Optional[list[Node[T]]]=...) -> None: ...
    def __iter__(self) -> Generator[Node[T], None, None]: ...
    def __getitem__(self, key: int) -> Optional[T]: ...
    def __len__(self) -> int: ...
    def getConnectionsOfNodeAtIndex(self, index: int) -> List[int]: ...
    def setNodeData(self, index: int, newValue: T) -> None: ...
    @staticmethod
    def createGraph(sizeX: int, sizeY: int) -> Graph[T]: ...
    def setNodesFromNodesList(self, nodes: list[Node[T]]) -> None: ...
    def setNodesFromValuesAndConnections(self, values: list[T], connectionsPointers: list[list[int]]) -> None: ...
    def connectionExistsFrom(self, indexA: int, indexB: int) -> bool: ...
    def depthFirstTraversal(self, nodeIndex: Optional[int]=...) -> Iterator[T]: ...
    def breadthFirstTraversal(self) -> Iterator[Optional[T]]: ...
    def removeLinkBetween(self, indexFrom: int, indexTo: int, bidirectional: bool=...) -> None: ...
    def addLinkBetween(self, indexFrom: int, indexTo: int, bidirectional: bool=...) -> None: ...



from typing import Generic, List, Optional, TypeVar

T = TypeVar("T")


class Node(Generic[T]):
    """A node of a graph. Stores data and its connections.

    Args:
        Generic ([type]): The type of data the node will store.
    """

    # This could easily be made into a Tuple rather than a whole new class but for debugging purposes, I chose to make this a class.

    data: Optional[
        T
    ]  # this node's data  - public because the graph may want to set its data again
    connections: list[
        int
    ]  # pointers of this node's connections   – public bc it needs to be read by the graph
    visited: bool = (
        False  # a helper variable to assist in traversals of a graph of nodes
    )

    def __init__(self, data: Optional[T], connections: list[int] = list[int]()) -> None:
        """Constructor for a binary tree node.

        Args:
            data (T): This node's data, e.g. 'Barry'.
            connections (list[int], optional): The list of pointers that are connections of this node. Defaults to an emply list[int]().
        """
        self.data = data
        self.connections = connections

    def __repr__(self) -> str:
        """Return a string representation of this object

        Returns:
            str: The string representation of the object, for example:
        ```
        "Alice -> (1, 4)"
        ```
        """
        # For example, "Bob -> [0, 4]"
        return str(
            str(self.data) + " -> [" + ", ".join(map(str, self.connections)) + "]"
        )

    def clone(self) -> "Node[T]":
        #  make sure to clone by value, not reference.
        connections: List[int] = [x for x in self.connections]
        return Node[T](self.data, connections)


#!python3.9

from typing import List, Optional, TypeVar, Generic

T = TypeVar("T")


class Stack(Generic[T]):
    __stack: List[Optional[T]] = list()
    __top = -1

    def pop(self) -> T:
        if not self.isEmpty():
            item = self.__stack[self.__top]
            self.__stack[self.__top] = None
            self.__top -= 1
            if item is not None:
                return (
                    item  # cannot be None because we checked whether self.__isEmpty()
                )
        raise IndexError("Stack empty")

    def peek(self) -> T:
        if not self.isEmpty():
            data = self.__stack[self.__top]
            if data is not None:
                return (
                    data  # cannot be None because we checked whether self.__isEmpty()
                )
        raise IndexError("Stack empty")

    def push(self, item: T) -> None:
        self.__top += 1
        if len(self.__stack) <= self.__top:
            self.__stack.append(item)
        else:
            self.__stack[self.__top] = item

    def isEmpty(self) -> bool:
        return self.__top == -1

    def __len__(self) -> int:
        return self.__top + 1


#!python3.9

from setuptools import setup  # type: ignore

setup(
    name="iron_datastructures",
    version="1.0",
    description="A set of useful data structures, including circular queue, binary tree/graph, linked list, and stack.",
    url="https://github.com/ntflix/iron_datastructures",
    author="Felix Weber",
    author_email="felix@iron59.co.uk",
    license="GNU GPLv3",
    package_data={
        "iron_datastructures": [
            "py.typed",
            "stack.pyi",
            "linked_list.pyi",
            "linked_list_node.pyi",
            "graph.pyi",
            "graph_node.pyi",
            "circular_queue.pyi",
            "_queue_position.pyi",
        ]
    },
    packages=["iron_datastructures"],
    zip_safe=False,
)



from datastructures.linked_list_node import LinkedListNode as LinkedListNode
from typing import Any, Generator, Optional, TypeVar

T = TypeVar('T')

class LinkedList:
    def __init__(self, nodes: list[T]=...) -> None: ...
    def __iter__(self) -> Generator[LinkedListNode[T], None, None]: ...
    def __getitem__(self, key: int) -> T: ...
    def __setitem__(self, key: int, newValue: T) -> Any: ...
    def setItemRecursively(self, index: int, newValue: T, currentNode: Optional[LinkedListNode[T]]=..., recursiveIndex: int=...) -> Optional[LinkedListNode[T]]: ...
    def toNode(self) -> Optional[LinkedListNode[T]]: ...
    def clone(self) -> LinkedList[T]: ...
    def toList(self) -> list[T]: ...
    def insertLinkedListAtBeginning(self, linkedList: LinkedList[T]) -> None: ...
    def insertNodeAtBeginning(self, nodeValues: LinkedListNode[T], clone: bool=...) -> None: ...
    def extendByLinkedList(self, linkedList: LinkedList[T]) -> None: ...
    def extendByNode(self, node: LinkedListNode[T], insertOnto: Optional[LinkedListNode[T]]=...) -> None: ...
    def removeNode(self, toRemove: T, node: Optional[LinkedListNode[T]]=...) -> None: ...
